@@ -18,7 +18,7 @@ from .math import extrapolation, backend
 from .physics import fluid, advect, diffuse
 
 # Classes
-from .math import Tensor, DType, Solve
+from .math import Shape, Tensor, DType, Solve
 from .geom import Geometry, Sphere, Box, Cuboid
 from .field import Grid, CenteredGrid, StaggeredGrid, mask, Noise, PointCloud, Scene, resample, GeometryMask, SoftGeometryMask, HardGeometryMask
 from .vis import Viewer
@@ -26,6 +26,7 @@ from .physics.fluid import Obstacle
 
 # Constants
 from .math import PI, INF, NAN
+from .math.extrapolation import PERIODIC, ZERO_GRADIENT
 
 # Functions
 from .math import (
