@@ -1,7 +1,7 @@
 import copy
 import warnings
 from numbers import Number
-from typing import TypeVar, Tuple, Set
+from typing import TypeVar, Tuple, Set, Dict, Union
 
 import dataclasses
 
@@ -15,12 +15,58 @@ from .magic import Sliceable, Shaped, Shapable, PhiTreeNode
 class MagicNotImplemented(Exception): pass
 
 
+def slice_(value, slices: Dict[str, Union[int, slice, str, tuple, list]]):
+    """
+    Slices a `Tensor` or `phi.math.magic.PhiTreeNode` along named dimensions.
+
+    See Also:
+        `unstack`.
+
+    Args:
+        value: `Tensor` or `phi.math.magic.PhiTreeNode`
+        slices: `dict` mapping dimension names to slices. A slice can be one of the following:
+
+            * An index (`int`)
+            * A range (`slice`)
+            * An item name (`str`)
+            * Multiple item names (comma-separated `str`)
+            * Multiple indices or item names (`tuple` or `list`)
+
+    Returns:
+        `Tensor` or `phi.math.magic.PhiTreeNode` of the same type as `value`.
+
+    Examples:
+        >>> math.slice([vec(x=0, y=1), vec(x=2, y=3)], {'vector': 'y'})
+        [1, 3]
+    """
+    if isinstance(value, (bool, Number)):
+        return value
+    if isinstance(value, tuple):
+        return tuple([slice_(v, slices) for v in value])
+    if isinstance(value, list):
+        return [slice_(v, slices) for v in value]
+    if isinstance(value, dict):
+        return {k: slice_(v, slices) for k, v in value.items()}
+    if isinstance(value, Shape):
+        raise NotImplementedError
+    if hasattr(value, '__getitem__'):
+        return value[slices]
+    if isinstance(value, PhiTreeNode):
+        attrs = {key: getattr(value, key) for key in value_attributes(value)}
+        new_attrs = {k: slice_(v, slices) for k, v in attrs.items()}
+        return copy_with(value, **new_attrs)
+    raise ValueError(f"value must be a PhiTreeNode but got {type(value)}")
+
+
 def unstack(value, dim: DimFilter):
     """
     Un-stacks a `Sliceable` along one or multiple dimensions.
 
     If multiple dimensions are given, the order of elements will be according to the dimension order in `dim`, i.e. elements along the last dimension will be neighbors in the returned `tuple`.
 
+    See Also:
+        `phi.math.slice`.
+
     Args:
         value: `phi.math.magic.Shapable`, such as `phi.math.Tensor`
         dim: Dimensions as `Shape` or comma-separated `str` or dimension type, i.e. `channel`, `spatial`, `instance`, `batch`.
@@ -42,7 +88,7 @@ def unstack(value, dim: DimFilter):
                 assert isinstance(result, tuple), f"__unstack__ must return a tuple but got {type(result)}"
                 assert all([isinstance(item, Sliceable) for item in result]), f"__unstack__ must return a tuple of Sliceable objects but not all items were sliceable in {result}"
                 return result
-        return tuple([value[{dims.name: i}] for i in range(dims.size)])
+        return tuple([slice_(value, {dims.name: i}) for i in range(dims.size)])
     else:  # multiple dimensions
         if hasattr(value, '__pack_dims__'):
             packed_dim = batch('_unstack')
@@ -253,6 +299,8 @@ def expand(value, *dims: Shape, **kwargs):
     Adds dimensions to a `Tensor` or tensor-like object by implicitly repeating the tensor values along the new dimensions.
     If `value` already contains any of the new dimensions, a size and type check is performed for these instead.
 
+    If any of `dims` varies along a dimension that is present neither in `value` nor on `dims`, it will also be added to `value`.
+
     This function replaces the usual `tile` / `repeat` functions of
     [NumPy](https://numpy.org/doc/stable/reference/generated/numpy.tile.html),
     [PyTorch](https://pytorch.org/docs/stable/tensors.html#torch.Tensor.repeat),
@@ -273,10 +321,11 @@ def expand(value, *dims: Shape, **kwargs):
         Same type as `value`.
     """
     dims = concat_shapes(*dims)
-    merge_shapes(value, dims.only(shape(value)))  # check that existing sizes match
+    combined = merge_shapes(value, dims)  # check that existing sizes match
     if not dims.without(shape(value)):  # no new dims to add
         if set(dims) == set(shape(value).only(dims)):  # sizes and item names might differ, though
             return value
+    dims &= combined.shape.without('dims')  # add missing non-uniform dims
     # --- First try __stack__
     if hasattr(value, '__expand__'):
         result = value.__expand__(dims, **kwargs)
