@@ -70,8 +70,18 @@ class Shape:
             for name, size in zip(names, sizes):
                 if size is not None and isinstance(size, Tensor):
                     assert size.rank > 0
-                    # for dim in size.shape.names:
-                    #     assert dim in self.names, f"Dimension {name} varies along {dim} but {dim} is not part of the Shape {self}"
+            for size, item_names in zip(self.sizes, self.item_names):
+                if item_names is not None:
+                    assert len(item_names) == size, f"Number of item names ({len(item_names)}) does not match size {size}"
+
+    def _check_is_valid_tensor_shape(self):
+        if DEBUG_CHECKS:
+            from ._tensors import Tensor
+            for name, size in zip(self.names, self.sizes):
+                if size is not None and isinstance(size, Tensor):
+                    assert size.rank > 0
+                    for dim in size.shape.names:
+                        assert dim in self.names, f"Dimension {name} varies along {dim} but {dim} is not part of the Shape {self}"
 
     def _to_dict(self, include_sizes=True):
         result = dict(names=self.names, types=self.types, item_names=self.item_names)
@@ -436,6 +446,17 @@ class Shape:
         """
         return self[[i for i, s in enumerate(self.sizes) if not _size_equal(s, 1)]]
 
+    @property
+    def singleton(self) -> 'Shape':
+        """
+        Filters this shape, returning only singleton dimensions as a new `Shape` object.
+        Dimensions are singleton if their size is exactly `1`.
+
+        Returns:
+            New `Shape` object
+        """
+        return self[[i for i, s in enumerate(self.sizes) if _size_equal(s, 1)]]
+
     def unstack(self, dim='dims') -> Tuple['Shape']:
         """
         Slices this `Shape` along a dimension.
@@ -1059,6 +1080,7 @@ class Shape:
                 if start < 0 and step > 0:
                     start += self.get_size(sel_dim)
                     assert start >= 0
+                stop = min(stop, self.get_size(sel_dim))
                 new_size = math.to_int64(math.ceil(math.wrap((stop - start) / step)))
                 if new_size.rank == 0:
                     new_size = int(new_size)  # NumPy array not allowed because not hashable
@@ -1232,6 +1254,7 @@ def _construct_shape(dim_type: str, prefix: str, *args, **dims):
             items = tuple([i.strip() for i in size.split(',')])
             size = len(items)
         elif isinstance(size, (tuple, list)):
+            assert all(isinstance(s, str) for s in size), f"Item names must all be of type 'str' but got '{size}'"
             items = tuple(size)
             size = len(items)
         elif isinstance(size, Shape):
