@@ -4,19 +4,17 @@ import sys
 import warnings
 from contextlib import contextmanager
 from threading import Thread
-from typing import Tuple, List, Dict, Callable
+from typing import Tuple, List, Dict
 
 from ._user_namespace import get_user_namespace, UserNamespace, DictNamespace
 from ._viewer import create_viewer, Viewer
-from ._vis_base import Control, value_range, Action, VisModel, Gui, \
-    PlottingLibrary, tensor_as_field
-from .. import math, field
-from ..field import SampledField, Scene, Field, PointCloud, Grid
+from ._vis_base import Control, value_range, Action, VisModel, Gui, PlottingLibrary, tensor_as_field, common_index, index_label, title_label
+from .. import math
+from ..field import SampledField, Scene, Field, PointCloud
 from ..field._scene import _slugify_filename
 from ..geom import Geometry, Box, embed
-from ..math import Tensor, layout, batch, Shape, spatial, channel
-from ..math._shape import parse_dim_names, parse_dim_order, DimFilter
-from ..math._tensors import Layout
+from ..math import Tensor, layout, batch, Shape, wrap, merge_shapes, EMPTY_SHAPE
+from ..math._shape import parse_dim_order, DimFilter
 
 
 def show(*model: VisModel or SampledField or tuple or list or Tensor or Geometry,
@@ -268,7 +266,7 @@ def get_current_figure():
     return LAST_FIGURE[0]
 
 
-def plot(*fields: SampledField or Tensor or Layout,
+def plot(*fields: SampledField or Tensor,
          lib: str or PlottingLibrary = None,
          row_dims: DimFilter = None,
          col_dims: DimFilter = batch,
@@ -279,7 +277,8 @@ def plot(*fields: SampledField or Tensor or Layout,
          same_scale=True,
          log_dims: str or tuple or list or Shape = '',
          show_color_bar=True,
-         color=None,
+         color: str or int or Tensor = None,
+         alpha: float or Tensor = 1.,
          frame_time=100,
          repeat=True):
     """
@@ -306,7 +305,12 @@ def plot(*fields: SampledField or Tensor or Layout,
             Can be given as a comma-separated `str`, a sequence of dimension names or a `Shape`.
             Use `'_'` to scale unnamed axes logarithmically, e.g. the y-axis of scalar functions.
         show_color_bar: Whether to display color bars for heat maps.
-        color: Tensor for line / marker colors.
+        color: `Tensor` of line / marker colors.
+            The color can be specified either as a cycle index (int tensor) or as a hex code (str tensor).
+            The color of different lines and markers can vary.
+        alpha: Opacity as `float` or `Tensor`.
+            This affects all elements, not only line plots.
+            Opacity can vary between lines and markers.
         animate: Time dimension to animate.
             If not present in the data, will produce a regular plot instead.
         overlay: Dimensions along which elements should be overlaid in the same subplot.
@@ -321,7 +325,9 @@ def plot(*fields: SampledField or Tensor or Layout,
 
         In case of an animation, a displayable animation object will be returned instead of a `Tensor`.
     """
-    nrows, ncols, fig_shape, positioning, indices = layout_sub_figures(math.layout(fields, batch('args')), row_dims, col_dims, animate, overlay, 0, 0, {}, {})
+    positioning = {}
+    indices: Dict[Tuple[int, int], List[dict]] = {}
+    nrows, ncols, fig_shape, reduced_shape = layout_sub_figures(layout(fields, batch('args')), row_dims, col_dims, animate, overlay, 0, 0, positioning, indices, {})
     animate = fig_shape.only(animate)
     fig_shape = fig_shape.without(animate)
     plots = default_plots() if lib is None else get_plots(lib)
@@ -342,19 +348,20 @@ def plot(*fields: SampledField or Tensor or Layout,
         title = {(row, col): title.rows[row].cols[col].native() for (row, col) in positioning}
     else:
         assert title is None, f"title must be a str or Tensor but got {title}"
-        title = {pos: ", ".join([i for dim, i in index.items() if isinstance(i, str)]) for pos, index in indices.items()}
+        title = {pos: title_label(common_index(*i, exclude=reduced_shape.singleton)) for pos, i in indices.items()}
     log_dims = parse_dim_order(log_dims) or ()
-    color = math.wrap(color)
+    color = layout_pytree_node(color, wrap_leaf=True)
+    alpha = layout_pytree_node(alpha, wrap_leaf=True)
     # --- animate or plot ---
     if fig_shape.volume == 1:
         figure, axes = plots.create_figure(size, nrows, ncols, subplots, title, log_dims)
         if animate:
             def plot_frame(frame: int):
                 for pos, fields in positioning.items():
-                    idx = indices[pos]
-                    for f in fields:
+                    for i, f in enumerate(fields):
+                        idx = indices[pos][i]
                         f = f[{animate.name: frame}]
-                        plots.plot(f, figure, axes[pos], subplots[pos], min_val, max_val, show_color_bar, color[idx])
+                        plots.plot(f, figure, axes[pos], subplots[pos], min_val, max_val, show_color_bar, color[idx], alpha[idx])
                 plots.finalize(figure)
             anim = plots.animate(figure, animate.size, plot_frame, frame_time, repeat)
             LAST_FIGURE[0] = anim
@@ -362,9 +369,9 @@ def plot(*fields: SampledField or Tensor or Layout,
             return anim
         else:
             for pos, fields in positioning.items():
-                idx = indices[pos]
-                for f in fields:
-                    plots.plot(f, figure, axes[pos], subplots[pos], min_val, max_val, show_color_bar, color[idx])
+                for i, f in enumerate(fields):
+                    idx = indices[pos][i]
+                    plots.plot(f, figure, axes[pos], subplots[pos], min_val, max_val, show_color_bar, color[idx], alpha[idx])
             plots.finalize(figure)
             LAST_FIGURE[0] = figure
             return layout(figure)
@@ -372,7 +379,17 @@ def plot(*fields: SampledField or Tensor or Layout,
         raise NotImplementedError(f"Figure batches not yet supported. Use rows and cols to reduce all batch dimensions. Not reduced. {fig_shape}")
 
 
-def layout_sub_figures(data: Tensor or Layout or SampledField,
+def layout_pytree_node(data, wrap_leaf=False):
+    if isinstance(data, tuple):
+        return layout(data, batch('tuple'))
+    elif isinstance(data, list):
+        return layout(data, batch('list'))
+    elif isinstance(data, dict):
+        return layout(data, batch('dict'))
+    return wrap(data) if wrap_leaf else data
+
+
+def layout_sub_figures(data: Tensor or SampledField,
                        row_dims: DimFilter,
                        col_dims: DimFilter,
                        animate: DimFilter,  # do not reduce these dims, has priority
@@ -380,48 +397,44 @@ def layout_sub_figures(data: Tensor or Layout or SampledField,
                        offset_row: int,
                        offset_col: int,
                        positioning: Dict[Tuple[int, int], List],
-                       base_index: Dict[str, int or str]) -> Tuple[int, int, Shape, dict, dict]:  # rows, cols
+                       indices: Dict[Tuple[int, int], List[dict]],
+                       base_index: Dict[str, int or str]) -> Tuple[int, int, Shape, Shape]:  # rows, cols
     if data is None:
         raise ValueError(f"Cannot layout figure for '{data}'")
-    if isinstance(data, list):
-        data = math.layout(data, batch('list'))
-    elif isinstance(data, tuple):
-        data = math.layout(data, batch('tuple'))
-    elif isinstance(data, dict):
-        data = math.layout(data, batch('dict'))
-    if isinstance(data, Layout):
+    data = layout_pytree_node(data)
+    if isinstance(data, Tensor) and data.dtype.kind == object:  # layout
         rows, cols = 0, 0
         non_reduced = math.EMPTY_SHAPE
-        indices = {}
-        dim0 = data.shape[0]
+        dim0 = reduced = data.shape[0]
         if dim0.only(overlay):
-            for d in data:  # overlay these fields
-                e_rows, e_cols, d_non_reduced, positioning, indices = layout_sub_figures(d, row_dims, col_dims, animate, overlay, offset_row, offset_col, positioning, base_index)
+            for overlay_index in dim0.only(overlay).meshgrid(names=True):  # overlay these fields
+                e_rows, e_cols, d_non_reduced, d_reduced = layout_sub_figures(data[overlay_index].native(), row_dims, col_dims, animate, overlay, offset_row, offset_col, positioning, indices, {**base_index, **overlay_index})
                 rows = max(rows, e_rows)
                 cols = max(cols, e_cols)
                 non_reduced &= d_non_reduced
+                reduced = merge_shapes(reduced, d_reduced, allow_varying_sizes=True)
         elif dim0.only(animate):
             data = math.stack(data.native(), dim0)
-            return layout_sub_figures(data, row_dims, col_dims, animate, overlay, offset_row, offset_col, positioning, base_index)
+            return layout_sub_figures(data, row_dims, col_dims, animate, overlay, offset_row, offset_col, positioning, indices, base_index)
         else:
             elements = data.unstack(dim0.name)
             for item_name, e in zip(dim0.get_item_names(dim0.name) or range(dim0.size), elements):
                 index = dict(base_index, **{dim0.name: item_name})
                 if dim0.only(row_dims):
-                    e_rows, e_cols, e_non_reduced, positioning, e_indices = layout_sub_figures(e.native(), row_dims, col_dims, animate, overlay, offset_row + rows, offset_col, positioning, index)
+                    e_rows, e_cols, e_non_reduced, e_reduced = layout_sub_figures(e.native(), row_dims, col_dims, animate, overlay, offset_row + rows, offset_col, positioning, indices, index)
                     rows += e_rows
                     cols = max(cols, e_cols)
                 elif dim0.only(col_dims):
-                    e_rows, e_cols, e_non_reduced, positioning, e_indices = layout_sub_figures(e.native(), row_dims, col_dims, animate, overlay, offset_row, offset_col + cols, positioning, index)
+                    e_rows, e_cols, e_non_reduced, e_reduced = layout_sub_figures(e.native(), row_dims, col_dims, animate, overlay, offset_row, offset_col + cols, positioning, indices, index)
                     cols += e_cols
                     rows = max(rows, e_rows)
                 else:
-                    e_rows, e_cols, e_non_reduced, positioning, e_indices = layout_sub_figures(e.native(), row_dims, col_dims, animate, overlay, offset_row, offset_col, positioning, index)
+                    e_rows, e_cols, e_non_reduced, e_reduced = layout_sub_figures(e.native(), row_dims, col_dims, animate, overlay, offset_row, offset_col, positioning, indices, index)
                     cols = max(cols, e_cols)
                     rows = max(rows, e_rows)
                 non_reduced &= e_non_reduced
-                indices.update(e_indices)
-        return rows, cols, non_reduced, positioning, indices
+                reduced = merge_shapes(reduced, e_reduced, allow_varying_sizes=True)
+        return rows, cols, non_reduced, reduced
     else:
         if isinstance(data, Tensor):
             data = tensor_as_field(data)
@@ -430,17 +443,16 @@ def layout_sub_figures(data: Tensor or Layout or SampledField,
         assert isinstance(data, Field), f"Cannot plot {type(data)}. Only tensors, geometries and fields can be plotted."
         overlay = data.shape.only(overlay)
         animate = data.shape.only(animate).without(overlay)
-        row_shape = batch(data).only(row_dims).without(animate).without(overlay)
-        col_shape = batch(data).only(col_dims).without(row_dims).without(animate).without(overlay)
+        row_shape = data.shape.only(row_dims).without(animate).without(overlay)
+        col_shape = data.shape.only(col_dims).without(row_dims).without(animate).without(overlay)
         non_reduced: Shape = batch(data).without(row_dims).without(col_dims) & animate
-        indices = {}
         for ri, r in enumerate(row_shape.meshgrid(names=True)):
             for ci, c in enumerate(col_shape.meshgrid(names=True)):
-                indices[(offset_row + ri, offset_col + ci)] = dict(base_index, **r, **c)
-                for o in overlay.meshgrid():
+                for o in overlay.meshgrid(names=True):
                     sub_data = data[r][c][o]
                     positioning.setdefault((offset_row + ri, offset_col + ci), []).append(sub_data)
-        return row_shape.volume, col_shape.volume, non_reduced, positioning, indices
+                    indices.setdefault((offset_row + ri, offset_col + ci), []).append(dict(base_index, **r, **c, **o))
+        return row_shape.volume, col_shape.volume, non_reduced, EMPTY_SHAPE
 
 
 def _space(fields: Tuple[Field, ...], ignore_dims: Shape) -> Box:
@@ -471,7 +483,7 @@ def overlay(*fields: SampledField or Tensor) -> Tensor:
     Returns:
         Plottable object
     """
-    return math.layout(fields, math.channel('overlay'))
+    return layout(fields, math.channel('overlay'))
 
 
 def write_image(path: str, figure=None, dpi=120.):
