@@ -395,7 +395,8 @@ class Recipe:
              min_val: float,
              max_val: float,
              show_color_bar: bool,
-             color: Tensor):
+             color: Tensor,
+             alpha: Tensor):
         raise NotImplementedError
 
     def __repr__(self):
@@ -430,7 +431,7 @@ def display_name(python_name: Any):
 def index_label(idx: dict) -> str or None:
     if len(idx) == 0:
         return None
-    if len(idx) == 1:
+    elif len(idx) == 1:
         return display_name(next(iter(idx.values())))
     else:
         number_unlabelled_dims = len([1 for k, v in idx.items() if isinstance(v, int)])
@@ -440,6 +441,25 @@ def index_label(idx: dict) -> str or None:
             return ", ".join(f'{k}={display_name(v)}' for k, v in idx.items())
 
 
+def title_label(idx: dict):
+    idx = {k: v for k, v in idx.items() if k not in ['tuple', 'list', 'dict', 'args'] or isinstance(v, str)}
+    if len(idx) == 0:
+        return None
+    elif len(idx) == 1:
+        for name, value in idx.items():
+            if isinstance(value, int):
+                return f"{display_name(name)} {display_name(value)}"
+            else:
+                return display_name(value)
+    else:
+        return index_label(idx)
+
+
+
+def common_index(*indices: dict, exclude=()):
+    return {k: v for k, v in indices[0].items() if k not in exclude and all(i[k] == v for i in indices)}
+
+
 def select_channel(value: SampledField or Tensor or tuple or list, channel: str or None):
     if isinstance(value, (tuple, list)):
         return [select_channel(v, channel) for v in value]
